# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import schema_pb2 as schema__pb2


class ImmuServiceStub(object):
    """immudb gRPC & REST service
    IMPORTANT: All get and safeget functions return base64-encoded keys and values, while all set and safeset functions expect base64-encoded inputs.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListUsers = channel.unary_unary(
                '/immudb.schema.ImmuService/ListUsers',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=schema__pb2.UserList.FromString,
                )
        self.CreateUser = channel.unary_unary(
                '/immudb.schema.ImmuService/CreateUser',
                request_serializer=schema__pb2.CreateUserRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.ChangePassword = channel.unary_unary(
                '/immudb.schema.ImmuService/ChangePassword',
                request_serializer=schema__pb2.ChangePasswordRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.UpdateAuthConfig = channel.unary_unary(
                '/immudb.schema.ImmuService/UpdateAuthConfig',
                request_serializer=schema__pb2.AuthConfig.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.UpdateMTLSConfig = channel.unary_unary(
                '/immudb.schema.ImmuService/UpdateMTLSConfig',
                request_serializer=schema__pb2.MTLSConfig.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.Login = channel.unary_unary(
                '/immudb.schema.ImmuService/Login',
                request_serializer=schema__pb2.LoginRequest.SerializeToString,
                response_deserializer=schema__pb2.LoginResponse.FromString,
                )
        self.Logout = channel.unary_unary(
                '/immudb.schema.ImmuService/Logout',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.Set = channel.unary_unary(
                '/immudb.schema.ImmuService/Set',
                request_serializer=schema__pb2.SetRequest.SerializeToString,
                response_deserializer=schema__pb2.TxMetadata.FromString,
                )
        self.VerifiableSet = channel.unary_unary(
                '/immudb.schema.ImmuService/VerifiableSet',
                request_serializer=schema__pb2.VerifiableSetRequest.SerializeToString,
                response_deserializer=schema__pb2.VerifiableTx.FromString,
                )
        self.Get = channel.unary_unary(
                '/immudb.schema.ImmuService/Get',
                request_serializer=schema__pb2.KeyRequest.SerializeToString,
                response_deserializer=schema__pb2.Entry.FromString,
                )
        self.VerifiableGet = channel.unary_unary(
                '/immudb.schema.ImmuService/VerifiableGet',
                request_serializer=schema__pb2.VerifiableGetRequest.SerializeToString,
                response_deserializer=schema__pb2.VerifiableEntry.FromString,
                )
        self.GetAll = channel.unary_unary(
                '/immudb.schema.ImmuService/GetAll',
                request_serializer=schema__pb2.KeyListRequest.SerializeToString,
                response_deserializer=schema__pb2.Entries.FromString,
                )
        self.ExecAll = channel.unary_unary(
                '/immudb.schema.ImmuService/ExecAll',
                request_serializer=schema__pb2.ExecAllRequest.SerializeToString,
                response_deserializer=schema__pb2.TxMetadata.FromString,
                )
        self.Scan = channel.unary_unary(
                '/immudb.schema.ImmuService/Scan',
                request_serializer=schema__pb2.ScanRequest.SerializeToString,
                response_deserializer=schema__pb2.Entries.FromString,
                )
        self.Count = channel.unary_unary(
                '/immudb.schema.ImmuService/Count',
                request_serializer=schema__pb2.KeyPrefix.SerializeToString,
                response_deserializer=schema__pb2.EntryCount.FromString,
                )
        self.CountAll = channel.unary_unary(
                '/immudb.schema.ImmuService/CountAll',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=schema__pb2.EntryCount.FromString,
                )
        self.TxById = channel.unary_unary(
                '/immudb.schema.ImmuService/TxById',
                request_serializer=schema__pb2.TxRequest.SerializeToString,
                response_deserializer=schema__pb2.Tx.FromString,
                )
        self.VerifiableTxById = channel.unary_unary(
                '/immudb.schema.ImmuService/VerifiableTxById',
                request_serializer=schema__pb2.VerifiableTxRequest.SerializeToString,
                response_deserializer=schema__pb2.VerifiableTx.FromString,
                )
        self.History = channel.unary_unary(
                '/immudb.schema.ImmuService/History',
                request_serializer=schema__pb2.HistoryRequest.SerializeToString,
                response_deserializer=schema__pb2.Entries.FromString,
                )
        self.Health = channel.unary_unary(
                '/immudb.schema.ImmuService/Health',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=schema__pb2.HealthResponse.FromString,
                )
        self.CurrentState = channel.unary_unary(
                '/immudb.schema.ImmuService/CurrentState',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=schema__pb2.ImmutableState.FromString,
                )
        self.SetReference = channel.unary_unary(
                '/immudb.schema.ImmuService/SetReference',
                request_serializer=schema__pb2.ReferenceRequest.SerializeToString,
                response_deserializer=schema__pb2.TxMetadata.FromString,
                )
        self.VerifiableSetReference = channel.unary_unary(
                '/immudb.schema.ImmuService/VerifiableSetReference',
                request_serializer=schema__pb2.VerifiableReferenceRequest.SerializeToString,
                response_deserializer=schema__pb2.VerifiableTx.FromString,
                )
        self.ZAdd = channel.unary_unary(
                '/immudb.schema.ImmuService/ZAdd',
                request_serializer=schema__pb2.ZAddRequest.SerializeToString,
                response_deserializer=schema__pb2.TxMetadata.FromString,
                )
        self.VerifiableZAdd = channel.unary_unary(
                '/immudb.schema.ImmuService/VerifiableZAdd',
                request_serializer=schema__pb2.VerifiableZAddRequest.SerializeToString,
                response_deserializer=schema__pb2.VerifiableTx.FromString,
                )
        self.ZScan = channel.unary_unary(
                '/immudb.schema.ImmuService/ZScan',
                request_serializer=schema__pb2.ZScanRequest.SerializeToString,
                response_deserializer=schema__pb2.ZEntries.FromString,
                )
        self.CreateDatabase = channel.unary_unary(
                '/immudb.schema.ImmuService/CreateDatabase',
                request_serializer=schema__pb2.Database.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.DatabaseList = channel.unary_unary(
                '/immudb.schema.ImmuService/DatabaseList',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=schema__pb2.DatabaseListResponse.FromString,
                )
        self.UseDatabase = channel.unary_unary(
                '/immudb.schema.ImmuService/UseDatabase',
                request_serializer=schema__pb2.Database.SerializeToString,
                response_deserializer=schema__pb2.UseDatabaseReply.FromString,
                )
        self.CleanIndex = channel.unary_unary(
                '/immudb.schema.ImmuService/CleanIndex',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.ChangePermission = channel.unary_unary(
                '/immudb.schema.ImmuService/ChangePermission',
                request_serializer=schema__pb2.ChangePermissionRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.SetActiveUser = channel.unary_unary(
                '/immudb.schema.ImmuService/SetActiveUser',
                request_serializer=schema__pb2.SetActiveUserRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class ImmuServiceServicer(object):
    """immudb gRPC & REST service
    IMPORTANT: All get and safeget functions return base64-encoded keys and values, while all set and safeset functions expect base64-encoded inputs.
    """

    def ListUsers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateUser(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ChangePassword(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateAuthConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateMTLSConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Login(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Logout(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Set(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def VerifiableSet(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Get(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def VerifiableGet(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAll(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecAll(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Scan(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Count(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CountAll(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TxById(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def VerifiableTxById(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def History(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Health(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CurrentState(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetReference(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def VerifiableSetReference(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ZAdd(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def VerifiableZAdd(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ZScan(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateDatabase(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DatabaseList(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UseDatabase(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CleanIndex(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ChangePermission(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetActiveUser(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ImmuServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ListUsers': grpc.unary_unary_rpc_method_handler(
                    servicer.ListUsers,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=schema__pb2.UserList.SerializeToString,
            ),
            'CreateUser': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateUser,
                    request_deserializer=schema__pb2.CreateUserRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'ChangePassword': grpc.unary_unary_rpc_method_handler(
                    servicer.ChangePassword,
                    request_deserializer=schema__pb2.ChangePasswordRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'UpdateAuthConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateAuthConfig,
                    request_deserializer=schema__pb2.AuthConfig.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'UpdateMTLSConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateMTLSConfig,
                    request_deserializer=schema__pb2.MTLSConfig.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'Login': grpc.unary_unary_rpc_method_handler(
                    servicer.Login,
                    request_deserializer=schema__pb2.LoginRequest.FromString,
                    response_serializer=schema__pb2.LoginResponse.SerializeToString,
            ),
            'Logout': grpc.unary_unary_rpc_method_handler(
                    servicer.Logout,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'Set': grpc.unary_unary_rpc_method_handler(
                    servicer.Set,
                    request_deserializer=schema__pb2.SetRequest.FromString,
                    response_serializer=schema__pb2.TxMetadata.SerializeToString,
            ),
            'VerifiableSet': grpc.unary_unary_rpc_method_handler(
                    servicer.VerifiableSet,
                    request_deserializer=schema__pb2.VerifiableSetRequest.FromString,
                    response_serializer=schema__pb2.VerifiableTx.SerializeToString,
            ),
            'Get': grpc.unary_unary_rpc_method_handler(
                    servicer.Get,
                    request_deserializer=schema__pb2.KeyRequest.FromString,
                    response_serializer=schema__pb2.Entry.SerializeToString,
            ),
            'VerifiableGet': grpc.unary_unary_rpc_method_handler(
                    servicer.VerifiableGet,
                    request_deserializer=schema__pb2.VerifiableGetRequest.FromString,
                    response_serializer=schema__pb2.VerifiableEntry.SerializeToString,
            ),
            'GetAll': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAll,
                    request_deserializer=schema__pb2.KeyListRequest.FromString,
                    response_serializer=schema__pb2.Entries.SerializeToString,
            ),
            'ExecAll': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecAll,
                    request_deserializer=schema__pb2.ExecAllRequest.FromString,
                    response_serializer=schema__pb2.TxMetadata.SerializeToString,
            ),
            'Scan': grpc.unary_unary_rpc_method_handler(
                    servicer.Scan,
                    request_deserializer=schema__pb2.ScanRequest.FromString,
                    response_serializer=schema__pb2.Entries.SerializeToString,
            ),
            'Count': grpc.unary_unary_rpc_method_handler(
                    servicer.Count,
                    request_deserializer=schema__pb2.KeyPrefix.FromString,
                    response_serializer=schema__pb2.EntryCount.SerializeToString,
            ),
            'CountAll': grpc.unary_unary_rpc_method_handler(
                    servicer.CountAll,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=schema__pb2.EntryCount.SerializeToString,
            ),
            'TxById': grpc.unary_unary_rpc_method_handler(
                    servicer.TxById,
                    request_deserializer=schema__pb2.TxRequest.FromString,
                    response_serializer=schema__pb2.Tx.SerializeToString,
            ),
            'VerifiableTxById': grpc.unary_unary_rpc_method_handler(
                    servicer.VerifiableTxById,
                    request_deserializer=schema__pb2.VerifiableTxRequest.FromString,
                    response_serializer=schema__pb2.VerifiableTx.SerializeToString,
            ),
            'History': grpc.unary_unary_rpc_method_handler(
                    servicer.History,
                    request_deserializer=schema__pb2.HistoryRequest.FromString,
                    response_serializer=schema__pb2.Entries.SerializeToString,
            ),
            'Health': grpc.unary_unary_rpc_method_handler(
                    servicer.Health,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=schema__pb2.HealthResponse.SerializeToString,
            ),
            'CurrentState': grpc.unary_unary_rpc_method_handler(
                    servicer.CurrentState,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=schema__pb2.ImmutableState.SerializeToString,
            ),
            'SetReference': grpc.unary_unary_rpc_method_handler(
                    servicer.SetReference,
                    request_deserializer=schema__pb2.ReferenceRequest.FromString,
                    response_serializer=schema__pb2.TxMetadata.SerializeToString,
            ),
            'VerifiableSetReference': grpc.unary_unary_rpc_method_handler(
                    servicer.VerifiableSetReference,
                    request_deserializer=schema__pb2.VerifiableReferenceRequest.FromString,
                    response_serializer=schema__pb2.VerifiableTx.SerializeToString,
            ),
            'ZAdd': grpc.unary_unary_rpc_method_handler(
                    servicer.ZAdd,
                    request_deserializer=schema__pb2.ZAddRequest.FromString,
                    response_serializer=schema__pb2.TxMetadata.SerializeToString,
            ),
            'VerifiableZAdd': grpc.unary_unary_rpc_method_handler(
                    servicer.VerifiableZAdd,
                    request_deserializer=schema__pb2.VerifiableZAddRequest.FromString,
                    response_serializer=schema__pb2.VerifiableTx.SerializeToString,
            ),
            'ZScan': grpc.unary_unary_rpc_method_handler(
                    servicer.ZScan,
                    request_deserializer=schema__pb2.ZScanRequest.FromString,
                    response_serializer=schema__pb2.ZEntries.SerializeToString,
            ),
            'CreateDatabase': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateDatabase,
                    request_deserializer=schema__pb2.Database.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'DatabaseList': grpc.unary_unary_rpc_method_handler(
                    servicer.DatabaseList,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=schema__pb2.DatabaseListResponse.SerializeToString,
            ),
            'UseDatabase': grpc.unary_unary_rpc_method_handler(
                    servicer.UseDatabase,
                    request_deserializer=schema__pb2.Database.FromString,
                    response_serializer=schema__pb2.UseDatabaseReply.SerializeToString,
            ),
            'CleanIndex': grpc.unary_unary_rpc_method_handler(
                    servicer.CleanIndex,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'ChangePermission': grpc.unary_unary_rpc_method_handler(
                    servicer.ChangePermission,
                    request_deserializer=schema__pb2.ChangePermissionRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'SetActiveUser': grpc.unary_unary_rpc_method_handler(
                    servicer.SetActiveUser,
                    request_deserializer=schema__pb2.SetActiveUserRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'immudb.schema.ImmuService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ImmuService(object):
    """immudb gRPC & REST service
    IMPORTANT: All get and safeget functions return base64-encoded keys and values, while all set and safeset functions expect base64-encoded inputs.
    """

    @staticmethod
    def ListUsers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/ListUsers',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            schema__pb2.UserList.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/CreateUser',
            schema__pb2.CreateUserRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ChangePassword(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/ChangePassword',
            schema__pb2.ChangePasswordRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateAuthConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/UpdateAuthConfig',
            schema__pb2.AuthConfig.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateMTLSConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/UpdateMTLSConfig',
            schema__pb2.MTLSConfig.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Login(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/Login',
            schema__pb2.LoginRequest.SerializeToString,
            schema__pb2.LoginResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Logout(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/Logout',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Set(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/Set',
            schema__pb2.SetRequest.SerializeToString,
            schema__pb2.TxMetadata.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def VerifiableSet(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/VerifiableSet',
            schema__pb2.VerifiableSetRequest.SerializeToString,
            schema__pb2.VerifiableTx.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Get(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/Get',
            schema__pb2.KeyRequest.SerializeToString,
            schema__pb2.Entry.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def VerifiableGet(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/VerifiableGet',
            schema__pb2.VerifiableGetRequest.SerializeToString,
            schema__pb2.VerifiableEntry.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetAll(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/GetAll',
            schema__pb2.KeyListRequest.SerializeToString,
            schema__pb2.Entries.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ExecAll(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/ExecAll',
            schema__pb2.ExecAllRequest.SerializeToString,
            schema__pb2.TxMetadata.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Scan(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/Scan',
            schema__pb2.ScanRequest.SerializeToString,
            schema__pb2.Entries.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Count(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/Count',
            schema__pb2.KeyPrefix.SerializeToString,
            schema__pb2.EntryCount.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CountAll(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/CountAll',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            schema__pb2.EntryCount.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TxById(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/TxById',
            schema__pb2.TxRequest.SerializeToString,
            schema__pb2.Tx.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def VerifiableTxById(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/VerifiableTxById',
            schema__pb2.VerifiableTxRequest.SerializeToString,
            schema__pb2.VerifiableTx.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def History(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/History',
            schema__pb2.HistoryRequest.SerializeToString,
            schema__pb2.Entries.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Health(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/Health',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            schema__pb2.HealthResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CurrentState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/CurrentState',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            schema__pb2.ImmutableState.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetReference(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/SetReference',
            schema__pb2.ReferenceRequest.SerializeToString,
            schema__pb2.TxMetadata.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def VerifiableSetReference(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/VerifiableSetReference',
            schema__pb2.VerifiableReferenceRequest.SerializeToString,
            schema__pb2.VerifiableTx.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ZAdd(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/ZAdd',
            schema__pb2.ZAddRequest.SerializeToString,
            schema__pb2.TxMetadata.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def VerifiableZAdd(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/VerifiableZAdd',
            schema__pb2.VerifiableZAddRequest.SerializeToString,
            schema__pb2.VerifiableTx.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ZScan(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/ZScan',
            schema__pb2.ZScanRequest.SerializeToString,
            schema__pb2.ZEntries.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateDatabase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/CreateDatabase',
            schema__pb2.Database.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DatabaseList(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/DatabaseList',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            schema__pb2.DatabaseListResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UseDatabase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/UseDatabase',
            schema__pb2.Database.SerializeToString,
            schema__pb2.UseDatabaseReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CleanIndex(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/CleanIndex',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ChangePermission(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/ChangePermission',
            schema__pb2.ChangePermissionRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetActiveUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/immudb.schema.ImmuService/SetActiveUser',
            schema__pb2.SetActiveUserRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



LEAF_PREFIX = b'\x00'
NODE_PREFIX = b'\x01'
ROOT_CACHE_PATH = "rootfile"

PERMISSION_SYS_ADMIN = 255
PERMISSION_ADMIN = 254
PERMISSION_NONE = 0
PERMISSION_R = 1
PERMISSION_RW = 2

SET_KEY_PREFIX=b'\x00'
SORTED_KEY_PREFIX=b'\x01'

PLAIN_VALUE_PREFIX=b'\x00'
REFERENCE_VALUE_PREFIX=b'\x01'

OLDEST_FIRST=False
NEWEST_FIRST=True




import struct

_SetSeparator=b"_~|IMMU|~_"

def wrap_zindex_ref(key: bytes, index) -> bytes:
	fmt=">{}sQB".format(len(key))
	if index!=None and index.index!=None:
		ret=struct.pack(fmt,key,index.index,1)
	else:
		ret=struct.pack(fmt,key,0,0)
	return ret

def unwrap_zindex_ref(value:bytes):
	l=len(value)
	fmt=">{}sQB".format(l-8-1)
	key, index, flag = struct.unpack(fmt,value)
	if flag==0:
		index=None
	return (key,index)

def build_set_key(key: bytes, zset: bytes, score: float, index) -> bytes:
	ret=_SetSeparator + zset + _SetSeparator + float64_2_bytes(score) + key
	ret=wrap_zindex_ref(ret,index)
	return ret

def float64_2_bytes(f:float)->bytes:
	return struct.pack(">d",f)
	



#!/usr/bin/env python3
SHA256LEN=32
import copy
import struct
import hashlib
from .constants import *

class printable:
    def __repr__(self)->str:
        ret="class {}\n".format(type(self).__name__)
        for k in self.__dict__:
            v=self.__dict__[k]
            if type(v)==bytes:
                ret+="\t{}: {}\n".format(k,list(v))
            elif hasattr(v,'__getitem__') and type(v[0])==bytes:
                ret+="\t{}:\n".format(k)
                for t in v:
                    ret+="\t- {}\n".format(list(t))
            else:
                ret+="\t{}: {}\n".format(k,v)
        return ret
    

class TXe(printable):
    def __init__(self):
        self.keyLen=None
        self.key=None
        self.ValueLen=None
        self.hValue=None
        self.vOff=None
    def SetKey(self, key: bytes):
        self.key=copy.copy(key)
        self.keyLen=len(key)
    def Digest(self):
        b=self.key+self.hValue
        return hashlib.sha256(b).digest()


        
class Tx(printable):
    def __init__(self):
        self.ID=None
        self.Ts=None
        self.BlTxID=None
        self.BlRoot=None
        self.PrevAlh=None
        self.nentries=None
        self.entries=None     
        self.htree=None
        self.Alh=None
        self.InnerHash=None
    def BuildHashTree(self):
        digests=[]
        for e in self.entries:
            digests.append(e.Digest())
        self.htree.BuildWith(digests)
    def CalcAlh(self):
        self.calcInnerhash()
        bi=struct.pack(">Q",self.ID)+self.PrevAlh+self.InnerHash
        self.Alh=hashlib.sha256(bi).digest()
    def calcInnerhash(self):
        bj=struct.pack(">QL",self.Ts,self.nentries)+self.eh()
        bj+=struct.pack(">Q",self.BlTxID)+self.BlRoot
        self.InnerHash=hashlib.sha256(bj).digest()
    def eh(self):
        return self.htree.root
    def Proof(self, key:bytes):
        kindex=None
        # find index of element holding given key 
        #kindex=next(k for k,v in enumerate(self.entries) if v.key==key)
        for k,v in enumerate(self.entries):
            if v.key==key:
                kindex=k
                break
        if kindex==None:
            raise LookupError
        return self.htree.InclusionProof(kindex)

def NewTxWithEntries(entries:list) -> Tx:
    tx=Tx()
    tx.ID=0
    tx.entries=entries
    tx.nentries=len(entries)
    tx.htree=HTree(len(entries))
    return tx
    
def TxFrom(stx) -> Tx:
    entries=[]
    for e in stx.entries:
        i=TXe()
        i.hValue=DigestFrom(e.hValue)
        i.vOff=e.vOff
        i.ValueLen=int(e.vLen)
        i.SetKey(e.key)
        entries.append(i)
    tx=NewTxWithEntries(entries)
    tx.ID = stx.metadata.id
    tx.PrevAlh = DigestFrom(stx.metadata.prevAlh)
    tx.Ts = stx.metadata.ts
    tx.BlTxID = stx.metadata.blTxId
    tx.BlRoot = DigestFrom(stx.metadata.blRoot)
    tx.BuildHashTree()
    tx.CalcAlh()
    return tx



class TxMetadata(printable):
    def __init__(self):
        self.iD       = None
        self.prevAlh  = None
        self.ts       = None
        self.nEntries = None
        self.eh       = None
        self.blTxID   = None
        self.blRoot   = None
    def alh(self):
        bi=struct.pack(">Q",self.iD)+self.prevAlh
        bj=struct.pack(">QL",int(self.ts),int(self.nEntries))
        bj=bj+self.eh+struct.pack(">Q",self.blTxID)+self.blRoot
        innerHash=hashlib.sha256(bj).digest()
        bi=bi+innerHash
        return hashlib.sha256(bi).digest()
    
def TxMetadataFrom(txmFrom):
    txm=TxMetadata()
    txm.iD       = txmFrom.id
    txm.prevAlh  = DigestFrom(txmFrom.prevAlh)
    txm.ts       = txmFrom.ts
    txm.nEntries = int(txmFrom.nentries)
    txm.eh       = DigestFrom(txmFrom.eH)
    txm.blTxID   = txmFrom.blTxId
    txm.blRoot   = DigestFrom(txmFrom.blRoot)
    return txm


    
class KV(printable):
    def __init__(self,key: bytes,value:bytes):
        self.key=key
        self.value=value
    def Digest(self):
        if self.value==None:
            valdigest=hashlib.sha256(b'').digest()
        else:
            valdigest=hashlib.sha256(self.value).digest()
        return hashlib.sha256(self.key+valdigest).digest()
        
def EncodeKey(key: bytes):
    return SET_KEY_PREFIX+key

def EncodeKV(key: bytes, value: bytes):
    return KV(SET_KEY_PREFIX+key,PLAIN_VALUE_PREFIX+value)

def EncodeReference(key:bytes, referencedKey: bytes, atTx: int):
    refVal=REFERENCE_VALUE_PREFIX+struct.pack(">Q",atTx)+SET_KEY_PREFIX+referencedKey
    return KV(SET_KEY_PREFIX+key,refVal)

def EncodeZAdd(zset:bytes, score:float, key: bytes, attx:int):
    ekey=SET_KEY_PREFIX+key
    zkey=SORTED_KEY_PREFIX
    zkey+=struct.pack(">Q",len(zset))+zset
    zkey+=struct.pack(">d",score)
    zkey+=struct.pack(">Q",len(ekey))+ekey
    zkey+=struct.pack(">Q",attx)
    return KV(zkey,None)


class LinearProof(printable):
    def __init__(self, sourceTxID:int, targetTxID:int, terms:list):
        self.sourceTxID=sourceTxID
        self.targetTxID=targetTxID
        self.terms=terms

def LinearProofFrom(lp)->LinearProof:
    return LinearProof(lp.sourceTxId, lp.TargetTxId, lp.terms)

def DigestFrom(slicedDigest: bytes)->bytes:
    d=copy.copy(slicedDigest[:SHA256LEN])
    return d

def DigestsFrom(slicedTerms: list) -> list:
    d=[copy.copy(i) for i in slicedTerms]
    return d
    


def VerifyInclusion(proof, digest: bytes, root) -> bool:
    if proof==None:
        return False
    leaf=LEAF_PREFIX+digest
    calcRoot = hashlib.sha256(leaf).digest()
    i=proof.leaf
    r=proof.width-1
    for t in proof.terms:
        b=NODE_PREFIX
        if i%2==0 and i!=r:
            b=b+calcRoot+t
        else:
            b=b+t+calcRoot
        calcRoot=hashlib.sha256(b).digest()
        i=i//2
        r=r//2
    return i==r and root==calcRoot

def leafFor(d:bytes)->bytes:
    b=LEAF_PREFIX+d
    return hashlib.sha256(b).digest()

def VerifyDualProof(proof, sourceTxID, targetTxID , sourceAlh, targetAlh):
    if (proof==None or
        proof.sourceTxMetadata==None or
        proof.targetTxMetadata==None or
        proof.sourceTxMetadata.iD != sourceTxID or
        proof.targetTxMetadata.iD != targetTxID):
            return False
    if proof.sourceTxMetadata.iD==0 or proof.sourceTxMetadata.iD > proof.targetTxMetadata.iD:
        return False
    if sourceAlh != proof.sourceTxMetadata.alh():
        return False
    if targetAlh != proof.targetTxMetadata.alh():
        return False
    if sourceTxID < proof.targetTxMetadata.blTxID and VerifyInclusionAHT( 
            proof.inclusionProof,
            sourceTxID,
            proof.targetTxMetadata.blTxID,
            leafFor(sourceAlh),
            proof.targetTxMetadata.blRoot)==False:
                return False
    if proof.sourceTxMetadata.blTxID > 0 and VerifyConsistency( 
            proof.consistencyProof,
            proof.sourceTxMetadata.blTxID,
            proof.targetTxMetadata.blTxID,
            proof.sourceTxMetadata.blRoot,
            proof.targetTxMetadata.blRoot)==False:
                return False
    if proof.targetTxMetadata.blTxID > 0 and VerifyLastInclusion( 
            proof.lastInclusionProof,
            proof.targetTxMetadata.blTxID,
            leafFor(proof.targetBlTxAlh),
            proof.targetTxMetadata.blRoot)==False:
                return False
    if sourceTxID < proof.targetTxMetadata.blTxID:
        ret=VerifyLinearProof(proof.linearProof, proof.targetTxMetadata.blTxID, targetTxID, proof.targetBlTxAlh, targetAlh) 
    else:
        ret=VerifyLinearProof(proof.linearProof, sourceTxID, targetTxID, sourceAlh, targetAlh)
    return ret

def VerifyInclusionAHT(iproof:list, i:int, j:int, iLeaf:bytes, jRoot:bytes) -> bool:
    if i>j or i==0 or i<j and len(iproof)==0:
        return False
    i1 = i - 1
    j1 = j - 1
    ciRoot = iLeaf
    for h in iproof:
        if i1%2 == 0 and i1 != j1:
            b=NODE_PREFIX+ciRoot+h
        else:
            b=NODE_PREFIX+h+ciRoot
        ciRoot = hashlib.sha256(b).digest()
        i1=i1>>1
        j1=j1>>1
    return jRoot==ciRoot

def VerifyConsistency(cproof:list, i:int, j:int, iRoot:bytes, jRoot:bytes)-> bool:
    if i > j or i == 0 or (i < j and len(cproof) == 0):
            return False
    if i == j and len(cproof) == 0:
            return iRoot == jRoot

    fn = i - 1
    sn = j - 1
    while fn%2 == 1:
        fn=fn>>1
        sn=sn>>1
    ciRoot, cjRoot = cproof[0], cproof[0]
    for h in cproof[1:]:
            if fn%2 == 1 or fn == sn:
                b=NODE_PREFIX+h+ciRoot
                ciRoot = hashlib.sha256(b).digest()
                b=NODE_PREFIX+h+cjRoot
                cjRoot = hashlib.sha256(b).digest()
                while fn%2 == 0 and fn != 0:
                    fn=fn>>1
                    sn=sn>>1
            else:
                b=NODE_PREFIX+cjRoot+h
                cjRoot=hashlib.sha256(b).digest()
            fn=fn>>1
            sn=sn>>1
    return iRoot == ciRoot and jRoot == cjRoot
    

def VerifyLastInclusion(iproof:list, i:int, leaf:bytes, root:bytes)->bool:
    if i==0:
        return False
    i1 = i - 1
    iroot = leaf
    for h in iproof:
        b=NODE_PREFIX+h+iroot
        iroot=hashlib.sha256(b).digest()
        i1 >>= 1
    return root==iroot
    
def VerifyLinearProof(proof , sourceTxID:int, targetTxID:int, sourceAlh:bytes, targetAlh:bytes) -> bool:
    if proof == None or proof.sourceTxID != sourceTxID or proof.targetTxID != targetTxID:
            return False

    if (proof.sourceTxID == 0 or proof.sourceTxID > proof.targetTxID or
            len(proof.terms) == 0 or sourceAlh != proof.terms[0]):
            return False

    calculatedAlh = proof.terms[0]
    for i in range(1,len(proof.terms)):
        bs=struct.pack(">Q",proof.sourceTxID+i)+calculatedAlh+proof.terms[i]
        calculatedAlh=hashlib.sha256(bs).digest()

    return targetAlh == calculatedAlh

class InclusionProof:
    def __init__(self):
        self.leaf=None
        self.width=None
        self.terms=[]

class DualProof:
    def __init__(self):
        self.sourceTxMetadata=None
        self.targetTxMetadata=None
        self.inclusionProof=None
        self.consistencyProof=None
        self.targetBlTxAlh=None
        self.lastInclusionProof=None
        self.linearProof=None
        
        
class HTree:
    def __init__(self, maxWidth:int):
        if maxWidth<1:
            return
       
        self.maxWidth=maxWidth
        lw=1
        while lw<maxWidth:
            lw=lw<<1
        height=(maxWidth-1).bit_length()+1
        self.levels=[None]*height
        for l in range(0,height):
            self.levels[l]=[None]*(lw>>l)

    def BuildWith(self, digests:list):
        if len(digests)>self.maxWidth:
            raise ErrMaxWidthExceeded
        if len(digests)==0:
            raise ErrIllegalArguments
        for i in range(0,len(digests)):
            leaf=LEAF_PREFIX+digests[i]
            self.levels[0][i]=hashlib.sha256(leaf).digest()
        l=0
        w=len(digests)
        while w>1:
            
            wn=0
            i=0
            while i+1<w:
                b=NODE_PREFIX+self.levels[l][i]+self.levels[l][i+1]
                self.levels[l+1][wn]=hashlib.sha256(b).digest()
                wn=wn+1
                i=i+2
            if w%2==1:
                self.levels[l+1][wn]=self.levels[l][w-1]
                wn=wn+1
            l+=1
            w=wn
        self.width=len(digests)
        self.root=self.levels[l][0]
    def InclusionProof(self, i):
        if i>=self.width:
            raise ErrIllegalArguments
        m=i
        n=self.width
        offset=0
        proof=InclusionProof()
        proof.leaf=i
        proof.width=self.width
        if self.width==1:
            return proof
        while True:
            d=(n-1).bit_length()
            k=1<<(d-1)
            if m<k:
                l,r=offset+k,offset+n-1
                n=k
            else:
                l,r=offset,offset+k-1
                m=m-k
                n=n-k
                offset=offset+k
            layer=(r-l).bit_length()
            index=int(l/(1<<layer))
            proof.terms.append(self.levels[layer][index])
            if n<1 or (n==1 and m==0):
                return proof
            
def InclusionProofFrom(iproof):
    h=InclusionProof()
    h.leaf=int(iproof.leaf)
    h.width=int(iproof.width)
    h.terms=DigestFrom(iproof.terms)
    return h

def DualProofFrom(dproof):
    dp=DualProof()
    dp.sourceTxMetadata=TxMetadataFrom(dproof.sourceTxMetadata)
    dp.targetTxMetadata=TxMetadataFrom(dproof.targetTxMetadata)
    dp.inclusionProof=DigestFrom(dproof.inclusionProof)
    dp.consistencyProof=DigestFrom(dproof.consistencyProof)
    dp.targetBlTxAlh=DigestFrom(dproof.targetBlTxAlh)
    dp.lastInclusionProof=DigestsFrom(dproof.lastInclusionProof)
    dp.linearProof=LinearProofFrom(dproof.linearProof) 
    return dp
    
    
    
